/*----------------------------------------------------------------------------

3D Physics Test Program - a cheesy test harness for 3D physics

by Chris Hecker for my Game Developer Magazine articles.  See my homepage
for more information.

NOTE: This is a hacked test program, not a nice example of Windows programming.
physics.cpp the only part of this you should look at!!!

This material is Copyright 1997 Chris Hecker, All Rights Reserved.
It's for you to read and learn from, not to put in your own articles
or books or on your website, etc.  Thank you.

Chris Hecker
checker@d6.com
http://www.d6.com/users/checker

*/

/*----------------------------------------------------------------------------

physics.cpp - This file implements the 3D physics simulator.

5/24/96 - checker

*/

#if 1       // windows

// it's nice that you have include windows.h to use the cross-platform OpenGL
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#else       // mac, and others?

#include <gl.h>
#include <glu.h>

#endif

#include <assert.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>

#include "physics.h"
#include "iface.h"
#include "linalg.h"

#include "mesh.h"

/*----------------------------------------------------------------------------

Globals

*/

real const WorldSize = r(100);
simulation_world *pWorld = 0;

/*----------------------------------------------------------------------------

various forces you can add to the system

@todo need to figure out units here so these numbers mean something

*/

int WorldSpringsActive = 0;
real const Kws = r(0.7);        // Hooke's spring constant
real const Kwd = r(0.1);        // damping constant
struct world_spring
{
    int unsigned BodyIndex;
    int unsigned VertexIndex;
    vector_3 Anchor;
    world_spring( int unsigned B, int unsigned V, vector_3 const &A ) :
            BodyIndex(B), VertexIndex(V), Anchor(A) {};
} aWorldSprings[1] =
{
    world_spring(simulation_world::NumberOfBodies-1,2,
                 vector_3(r(1.5),r(0),r(0))),
};
int unsigned const NumberOfWorldSprings =
    sizeof(aWorldSprings)/sizeof(aWorldSprings[0]);

int RotateHeadActive = 0;
int HeadBobbingActive = 1;
int StabilizersActive = 1;

int GravityActive = 1;
vector_3 Gravity(r(0),r(0),r(-10));

int DampingActive = 0;
// real const NoKdl = r(0.002);        // the no-damping linear damping factor
// real const NoKda = r(0.001);       // the no-damping angular damping factor
real const NoKdl = r(0.008);        // the no-damping linear damping factor
real const NoKda = r(0.004);       // the no-damping angular damping factor
real const Kdl = r(0.04);           // linear damping factor
real const Kda = r(0.01);           // angular damping factor

/*----------------------------------------------------------------------------
  
Init & Run

*/

void Initialize( void )
{
    srand(time(0));

    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    //glPolygonMode(GL_BACK,GL_LINE);
    glShadeModel(GL_FLAT);

    float aLightPosition[4] = { -1.0f, 1.0f, 1.0f, 0.0f };
    float aAmbient[4] = { 0.1f, 0.1f, 0.1f, 1.0f };
    glLightfv(GL_LIGHT0,GL_POSITION,aLightPosition);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT,aAmbient);
    glEnable(GL_LIGHT0);
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);

    glMatrixMode(GL_PROJECTION);
    glViewport(0,0,WindowWidth,WindowHeight);
    gluPerspective(45.0,float(WindowWidth)/float(WindowHeight),0.1,100.0);

    glMatrixMode(GL_MODELVIEW);

    pWorld = new simulation_world(WorldSize,WorldSize,WorldSize);
    assert(pWorld);
}

void Run( void )
{
    assert(pWorld);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    static double LastTime = GetTime();

#define FIXEDTIMESTEP 0
    // our Euler integrator sucks
    real const MaxTimeStep = r(0.0075);
    
#if FIXEDTIMESTEP
    real Time = LastTime + MaxTimeStep;
#else
    double Time = GetTime();
    if((Time - LastTime) > r(10)*MaxTimeStep)
    {
        LastTime = Time - MaxTimeStep;
    }
#endif

    while(LastTime < Time)
    {
        double DeltaTime = Time - LastTime;
        if(DeltaTime > MaxTimeStep)
        {
            DeltaTime = MaxTimeStep;
        }
        
        pWorld->Simulate(float(DeltaTime),float(LastTime));
        LastTime += DeltaTime;
    }
    LastTime = Time;

    pWorld->Render();
}

void Uninitialize( void )
{
    delete pWorld;
}

/*----------------------------------------------------------------------------

Toggles

*/

int ToggleRotateHead( void )
{
    return RotateHeadActive = RotateHeadActive ? 0 : 1;
}

int ToggleWorldSprings( void )
{
    return WorldSpringsActive = WorldSpringsActive ? 0 : 1;
}

int ToggleStabilizers( void )
{
    return StabilizersActive = StabilizersActive ? 0 : 1;
}

int ToggleGravity( void )
{
    return GravityActive = GravityActive ? 0 : 1;
}

int ToggleHeadBobbing( void )
{
    return HeadBobbingActive = HeadBobbingActive ? 0 : 1;
}

int ToggleDamping( void )
{
    return DampingActive = DampingActive ? 0 : 1;
}

/*----------------------------------------------------------------------------

utilities

*/

real GenerateUnitRandomReal( void )
{
    real Random = ((real)rand())/((real)RAND_MAX);
    assert((Random >= r(0)) && (Random <= r(1)));
    return Random;
}
        
real GenerateReasonableRandomReal( void )
{
    return r(0.1) + r(0.2)*GenerateUnitRandomReal();
}

real HeadRadius = r(0.7);
real HairRadius = r(0.15);
real JointLengthExtra = r(0.1);

void simulation_world::InitializeBodies( void )
{
    // first, the kinematically controlled head
    int BodyIndex = 0;
    int Mesh = 0;
    int Counter;

    rigid_body &Body = aBodies[BodyIndex];

    Body.OneOverMass = r(0);
    Body.InverseBodyInertiaTensor(0,0) = r(0);
    Body.InverseBodyInertiaTensor(1,1) = r(0);
    Body.InverseBodyInertiaTensor(2,2) = r(0);

    // @todo need to init first configuration
    Body.aConfigurations[0].InverseWorldInertiaTensor = Body.InverseBodyInertiaTensor;

    Body.CoefficientOfRestitution = r(1);

    Body.IsKinematic = true;

    // Body.aConfigurations[0].CMPosition;

    // initialize geometric quantities
    // we'll use the body index+1 for the display list id

    glNewList(BodyIndex+1,GL_COMPILE);
        for(Mesh = 0;Mesh < 2;Mesh++)
        {
            glShadeModel(GL_SMOOTH);
            glPushMatrix();
            glTranslatef(-0.7,0,0);
            real const s = 0.37;
            glScalef(s,s,s);
            glTranslatef(0.f,0.f,-2.104f);
            glRotatef(-90.f,0.f,0.f,1.f);
            glTranslatef(-aMeshes[Mesh].aCenter[0],
                         -aMeshes[Mesh].aCenter[1],
                         -aMeshes[Mesh].aCenter[2]);            
            glColor3fv(aMeshes[Mesh].aColor);
            glBegin(GL_TRIANGLES);
            for(Counter = 0;Counter < aMeshes[Mesh].NumberOfTriangles;Counter++)
            {
//                glNormal3fv(aMeshes[Mesh].aNormals[Counter]);
                for(int i = 0;i < 3;i++)
                {
                    int Index = aMeshes[Mesh].aIndices[Counter][i];
                    glNormal3fv(aMeshes[Mesh].aFaceVertexNormals[Counter][i]);
                    glVertex3fv(aMeshes[Mesh].aVertices[Index]);
                }
            }
            glEnd();
            glPopMatrix();
        }
    glEndList();

    Body.NumberOfBoundingVertices = 0;

    for(BodyIndex = 1;BodyIndex < NumberOfBodies;BodyIndex++)
    {
        // initialize the ponytail rigid bodies

        rigid_body &Body = aBodies[BodyIndex];

        // 1/2 the dimensions of the box
        real dX2 = HairRadius;
        real dY2 = HairRadius;
        real dZ2 = HairRadius; 

        real Density = r(0.4);
        real Mass = r(8)*Density*dX2*dY2*dZ2;
        assert(Mass > r(0));

        Body.OneOverMass = r(1)/Mass;
        Body.InverseBodyInertiaTensor(0,0) = r(3)/(Mass*(dY2*dY2+dZ2*dZ2));
        Body.InverseBodyInertiaTensor(1,1) = r(3)/(Mass*(dX2*dX2+dZ2*dZ2));
        Body.InverseBodyInertiaTensor(2,2) = r(3)/(Mass*(dX2*dX2+dY2*dY2));

        // @todo need to init first configuration
        Body.aConfigurations[0].InverseWorldInertiaTensor = Body.InverseBodyInertiaTensor;

        Body.CoefficientOfRestitution = r(1);

        Body.IsKinematic = false;

        Body.aConfigurations[0].CMPosition =
            vector_3(HeadRadius + JointLengthExtra +
                     HairRadius + JointLengthExtra +
                     (HairRadius + JointLengthExtra)*r(2)*r(BodyIndex-1),
                     r(0),r(0));

        // initialize geometric quantities
        // we'll use the body index+1 for the display list id

        glNewList(BodyIndex+1,GL_COMPILE);
            glPushMatrix();
            glRotatef(-90.f,0.f,1.f,0.f);
            real const s = 0.37;
            glScalef(s,s,s);
            int MeshEnd;
            if(BodyIndex < (NumberOfBodies-1))
            {
                glShadeModel(GL_SMOOTH);
                Mesh = 2;
                MeshEnd = 3;
                real s = 1.f - 0.3f*real(BodyIndex-1)/real(NumberOfBodies - 3);

                glScalef(s,s,1.0f);
                glRotatef(35.02289492520142,0,1,0);
            }
            else
            {
                glShadeModel(GL_FLAT);
                Mesh = 3;
                MeshEnd = 5;
                glTranslatef(-0.254,0,0.5);
            }
            glRotatef(-90.f,0.f,0.f,1.f);
            for(;Mesh < MeshEnd;Mesh++)
            {
                glPushMatrix();
                glTranslatef(-aMeshes[Mesh].aCenter[0],
                             -aMeshes[Mesh].aCenter[1],
                             -aMeshes[Mesh].aCenter[2]);            
                glColor3fv(aMeshes[Mesh].aColor);
                glBegin(GL_TRIANGLES);
                for(Counter = 0;Counter < aMeshes[Mesh].NumberOfTriangles;Counter++)
                {
                    for(int i = 0;i < 3;i++)
                    {
                        int Index = aMeshes[Mesh].aIndices[Counter][i];
                        glNormal3fv(aMeshes[Mesh].aFaceVertexNormals[Counter][i]);
                        glVertex3fv(aMeshes[Mesh].aVertices[Index]);
                    }
#if 0
                        glNormal3fv(aMeshes[Mesh].aFaceNormals[Counter]);
                        glVertex3fv(aMeshes[Mesh].aVertices[aMeshes[Mesh].aIndices[Counter][0]]);
                        glVertex3fv(aMeshes[Mesh].aVertices[aMeshes[Mesh].aIndices[Counter][1]]);
                        glVertex3fv(aMeshes[Mesh].aVertices[aMeshes[Mesh].aIndices[Counter][2]]);
#endif
                }
                glEnd();
                glPopMatrix();
            }
            glPopMatrix();
        glEndList();

        // generate the body-space bounding volume vertices

        assert(Body.MaxNumberOfBoundingVertices > 8);
        Body.NumberOfBoundingVertices = 8;
        Body.aBodyBoundingVertices[0] = vector_3( dX2, dY2, dZ2);
        Body.aBodyBoundingVertices[1] = vector_3( dX2, dY2,-dZ2);
        Body.aBodyBoundingVertices[2] = vector_3( dX2,-dY2, dZ2);
        Body.aBodyBoundingVertices[3] = vector_3( dX2,-dY2,-dZ2);
        Body.aBodyBoundingVertices[4] = vector_3(-dX2, dY2, dZ2);
        Body.aBodyBoundingVertices[5] = vector_3(-dX2, dY2,-dZ2);
        Body.aBodyBoundingVertices[6] = vector_3(-dX2,-dY2, dZ2);
        Body.aBodyBoundingVertices[7] = vector_3(-dX2,-dY2,-dZ2);
    }
}

void simulation_world::InitializeJoints( void )
{
    for(int i = 0;i < NumberOfJoints;i++)
    {
        spherical_joint &Joint = aJoints[i];
        // @todo inner and outer are a bit confusing here...from joint or body?
        spherical_joint::joint_definition &Inner = Joint.aJointDefinitions[0];
        spherical_joint::joint_definition &Outer = Joint.aJointDefinitions[1];

        Inner.BodyIndex = i;
        if(i == 0)
        {
            Inner.ToLocalJoint = vector_3(HeadRadius + JointLengthExtra,
                                          r(0),r(0));
        }
        else
        {
            Inner.ToLocalJoint = vector_3(HairRadius + JointLengthExtra,
                                          r(0),r(0));
        }

        Outer.BodyIndex = i + 1;
        Outer.ToLocalJoint = vector_3(-(HairRadius + JointLengthExtra),
                                      r(0),r(0));
    }
}

/*----------------------------------------------------------------------------

simulation_world ctor

*/

simulation_world::simulation_world(
    real WorldXLength, real WorldYLength, real WorldZLength ) :
    SourceConfigurationIndex(0), TargetConfigurationIndex(1)
{
    InitializeBodies();
    InitializeJoints();

    // initialize walls
    aWalls[0].Normal = vector_3(r(0),r(-1),r(0));
    aWalls[0].d = r(WorldYLength/2);
    aWalls[1].Normal = vector_3(r(0),r(1),r(0));
    aWalls[1].d = r(WorldYLength/2);

    aWalls[2].Normal = vector_3(r(-1),r(0),r(0));
    aWalls[2].d = r(WorldXLength/2);
    aWalls[3].Normal = vector_3(r(1),r(0),r(0));
    aWalls[3].d = r(WorldXLength/2);

    aWalls[4].Normal = vector_3(r(0),r(0),r(-1));
    aWalls[4].d = r(WorldZLength/2);
    aWalls[5].Normal = vector_3(r(0),r(0),r(1));
    aWalls[5].d = r(WorldZLength/2);

    // calculate initial bounding volume positions
    CalculateVertices(0);
}

/*----------------------------------------------------------------------------

simulation_world dtor

*/

simulation_world::~simulation_world( void )
{

}

/*----------------------------------------------------------------------------

Render - render the source configurations

*/

void simulation_world::Render( void )
{
    int Counter;


    // draw bodies
#if 1
    glEnable(GL_LIGHTING);
    for(Counter = 0;Counter < NumberOfBodies;Counter++)
    {
        matrix_3x3 &Orientation = aBodies[Counter].
                aConfigurations[SourceConfigurationIndex].Orientation;
        vector_3 &CMPosition = aBodies[Counter].
                aConfigurations[SourceConfigurationIndex].CMPosition;
        float aM[16];

        glPushMatrix();
            CreateOpenGLTransform(Orientation,CMPosition,aM);
            glMultMatrixf(aM);
            glCallList(Counter+1);
        glPopMatrix();
    }
    glDisable(GL_LIGHTING);

#endif
    
    // draw joints
    glBegin(GL_LINES);
    for(Counter = 0;Counter < NumberOfJoints;Counter++)
    {
        static float aJointColors[2][3] = {{1.f,0.f,0.f},{0.f,1.f,0.f}};
        for(int i = 0;i<2;i++)
        {
            spherical_joint::joint_definition &Joint =
                aJoints[Counter].aJointDefinitions[i];
            rigid_body::configuration &Configuration =
                aBodies[Joint.BodyIndex].
                aConfigurations[SourceConfigurationIndex];
            glColor3fv(aJointColors[i]);
            glVertex3fv(Configuration.CMPosition.GetElements());
            glVertex3fv((Configuration.CMPosition + 
                         Configuration.Orientation*Joint.ToLocalJoint).
                        GetElements());
        }
    }
    glEnd();

#if 0
    // draw the bounding volume points
    glBegin(GL_POINTS);
        glColor3f(1.0f,1.0f,1.0f);
        for(Counter = 0;Counter < NumberOfBodies;Counter++)
        {
            rigid_body::configuration &Configuration = 
                aBodies[Counter].aConfigurations[SourceConfigurationIndex];

            for(int unsigned i = 0;i < Body.NumberOfBoundingVertices;i++)
            {
                glVertex3fv(Configuration.aBoundingVertices[i].GetElements());
            }
        }
    glEnd();
#endif

    // draw springs

    if(WorldSpringsActive)
    {
        glBegin(GL_LINES);
            for(int i = 0;i < NumberOfWorldSprings;i++)
            {
                world_spring &Spring = aWorldSprings[i];
                rigid_body::configuration &Configuration =
                        aBodies[Spring.BodyIndex].
                        aConfigurations[SourceConfigurationIndex];
    
                glColor3f(0.0f,0.0f,0.0f);
                glVertex3fv(Configuration.aBoundingVertices[Spring.VertexIndex].GetElements());
                glVertex3fv(Spring.Anchor.GetElements());
            }
        glEnd();
    }

    // draw walls...I didn't bother implementing the cool clipper thing
    // like in the 2D sample
    assert(NumberOfWalls == 6);     // @todo sort of assert it's a cube

    glColor3f(1.0f,1.0f,1.0f);
    // do a big linestrip to get most of edges
    glBegin(GL_LINE_STRIP);
        glVertex3f(-WorldSize/r(2),-WorldSize/r(2),-WorldSize/r(2));
        glVertex3f(-WorldSize/r(2),-WorldSize/r(2), WorldSize/r(2));
        glVertex3f(-WorldSize/r(2), WorldSize/r(2), WorldSize/r(2));
        glVertex3f( WorldSize/r(2), WorldSize/r(2), WorldSize/r(2));
        glVertex3f( WorldSize/r(2),-WorldSize/r(2), WorldSize/r(2));
        glVertex3f( WorldSize/r(2),-WorldSize/r(2),-WorldSize/r(2));
    glEnd();
    // fill in the stragglers
    glBegin(GL_LINES);
        glVertex3f( WorldSize/r(2),-WorldSize/r(2), WorldSize/r(2));
        glVertex3f(-WorldSize/r(2),-WorldSize/r(2), WorldSize/r(2));

        glVertex3f( WorldSize/r(2), WorldSize/r(2), WorldSize/r(2));
        glVertex3f( WorldSize/r(2), WorldSize/r(2),-WorldSize/r(2));

        glVertex3f(-WorldSize/r(2), WorldSize/r(2), WorldSize/r(2));
        glVertex3f(-WorldSize/r(2), WorldSize/r(2),-WorldSize/r(2));
    glEnd();
    
    // draw floor
    glDisable(GL_CULL_FACE);
    glBegin(GL_QUADS);
        glColor3f(0.5f,0.5f,0.5f);
        glVertex3f( WorldSize/r(2), WorldSize/r(2),-WorldSize/r(2));
        glVertex3f(-WorldSize/r(2), WorldSize/r(2),-WorldSize/r(2));
        glVertex3f(-WorldSize/r(2),-WorldSize/r(2),-WorldSize/r(2));
        glVertex3f( WorldSize/r(2),-WorldSize/r(2),-WorldSize/r(2));
    glEnd();
    glEnable(GL_CULL_FACE);
}

/*----------------------------------------------------------------------------

Simulate - Integrate forward DeltaTime seconds.

@todo do I need to store the last simulation time?

*/

void simulation_world::Simulate( real DeltaTime, real CurrentTime )
{
    real TargetTime = CurrentTime + DeltaTime;

    while(CurrentTime < TargetTime)
    {
        ComputeForces(TargetTime,SourceConfigurationIndex);
        ComputeJointForces(SourceConfigurationIndex);

        Integrate(TargetTime-CurrentTime);

        CalculateVertices(TargetConfigurationIndex);

        CheckForCollisions(TargetConfigurationIndex);

        if(CollisionState == Penetrating)
        {
            // we simulated too far, so subdivide time and try again
            TargetTime = (CurrentTime + TargetTime) / r(2);

            // blow up if we aren't moving forward each step, which is
            // probably caused by interpenetration at the frame start

            assert(fabs(TargetTime - CurrentTime) > Epsilon);
        }
        else
        {
            // either colliding or clear

            if(CollisionState == Colliding)
            {
                // @todo handle multiple simultaneous collisions

                int Counter = 0;
                do
                {
                    ResolveCollisions(TargetConfigurationIndex);
                    Counter++;
                } while((CheckForCollisions(TargetConfigurationIndex) ==
                         Colliding) && (Counter < 100));

                assert(Counter < 100);
            }

            // we made a successful step, so swap configurations
            // to "save" the data for the next step
            
            CurrentTime = TargetTime;
            TargetTime = DeltaTime;

            SourceConfigurationIndex = SourceConfigurationIndex ? 0 : 1;
            TargetConfigurationIndex = TargetConfigurationIndex ? 0 : 1;
        }
    }
}

/*----------------------------------------------------------------------------

ComputeForces - compute forces for gravity, spring, etc.

*/

void simulation_world::ComputeForces( real Time, int ConfigurationIndex )
{
    int Counter;

    {
        // @todo if(!Body.IsKinematic)
        rigid_body &Body = aBodies[0];
        rigid_body::configuration &Configuration =
            Body.aConfigurations[ConfigurationIndex];

        if(HeadBobbingActive)
        {
            // circle equations
            // p = { cos(s*t), sin(s*t) }
            // p' = { -s*sin(s*t), s*cos(s*t) }
            // p'' = { -s*s*cos(s*t), -s*s*sin(s*t) }
            real sx = r(3), sy = r(6);
            real tx = sx*Time, ty = sy*Time;
            real sin_tx = r(sin(tx)), cos_tx = r(cos(tx));
            real sin_ty = r(sin(ty)), cos_ty = r(cos(ty));
            real ax = r(0.4), ay = ax;
            int x_index = 1, y_index = 2;

            real px = ax*cos_tx - Configuration.CMPosition(x_index);
            real px_dot = -ax*sx*sin_tx - Configuration.CMVelocity(x_index);
            real px_ddot = -ax*sx*sx*cos_tx;

            real py = ay*sin_ty - Configuration.CMPosition(y_index);
            real py_dot = ay*sy*cos_ty - Configuration.CMVelocity(y_index);
            real py_ddot = -ay*sy*sy*sin_ty;

            real OneOverTau = r(1) / r(0.25);
            real Alpha = r(2) * OneOverTau;
            real Beta = OneOverTau * OneOverTau;

            real x = px_ddot + Alpha * px_dot + Beta * px;
            real y = py_ddot + Alpha * py_dot + Beta * py;
            Configuration.CMAcceleration = vector_3();
            Configuration.CMAcceleration(x_index) = x;
            Configuration.CMAcceleration(y_index) = y;

        }
        else
        {
            Configuration.CMAcceleration = r(-1.0)*Configuration.CMVelocity;
        }

        real const Velocity = Configuration.AngularVelocity(2);
        real const MaxVelocity = PI/r(1);
        real const MaxAcceleration = (PI/r(4))/r(0.25);
        real TargetAcceleration;

        if(RotateHeadActive)
        {
            TargetAcceleration = MaxVelocity - Velocity;
        }
        else
        {
            /*
              So, we want the head to reorient itself to its default
              orientation.  We create a PD controller that will drive
              the head's y axis' angle with the world's x axis to PI/2.

              */
                        
            matrix_3x3 &O = Configuration.Orientation;
            real theta = PI/r(2) - (real)acos(O(0,1));    // head_y . world_x
            
            real OneOverTau = r(1) / r(0.25);
            real Alpha = r(2) * OneOverTau;
            real Beta = OneOverTau * OneOverTau;

            TargetAcceleration = -Alpha * Velocity + Beta * theta;
        }


        real Acceleration = ContinuousMaxMagnitude(TargetAcceleration,
                                                   MaxAcceleration);
        
        Configuration.AngularAcceleration = vector_3(0,0,Acceleration);
    }

    for(Counter = 0;Counter < NumberOfBodies;Counter++)
    {
        // @todo if(!Body.IsKinematic)
        rigid_body &Body = aBodies[Counter];
        rigid_body::configuration &Configuration =
            Body.aConfigurations[ConfigurationIndex];

        // clear forces

        Configuration.Torque = vector_3();
        Configuration.CMForce = vector_3();

        if(GravityActive && (Body.OneOverMass != r(0)))
        {
            Configuration.CMForce += Gravity / Body.OneOverMass; 
        }

        if(DampingActive)
        {
            Configuration.CMForce += -Kdl * Configuration.CMVelocity;
            Configuration.Torque += -Kda * Configuration.AngularVelocity;
        }
        else
        {
            // there's always a little damping because our integrator sucks
            Configuration.CMForce += -NoKdl * Configuration.CMVelocity;
            Configuration.Torque += -NoKda * Configuration.AngularVelocity;
        }
    }

    if(WorldSpringsActive)
    {
        for(int i = 0;i < NumberOfWorldSprings;i++)
        {
            world_spring &SpringStructure = aWorldSprings[i];
            
            rigid_body &Body = aBodies[SpringStructure.BodyIndex];
            rigid_body::configuration &Configuration =
                Body.aConfigurations[ConfigurationIndex];
            
            vector_3 Position = Configuration.aBoundingVertices[
                SpringStructure.VertexIndex];
            vector_3 U = Position - Configuration.CMPosition;
            vector_3 VU = Configuration.CMVelocity +
                CrossProduct(Configuration.AngularVelocity,U);
            
            vector_3 Spring = -Kws * (Position - SpringStructure.Anchor);
            // project velocity onto spring to get damping vector
            // this is basically a Gram-Schmidt projection
            vector_3 DampingForce =
                -Kwd*(DotProduct(VU,Spring)/DotProduct(Spring,Spring)) * Spring;
            
            Spring += DampingForce;
            
            Configuration.CMForce += Spring;
            Configuration.Torque += CrossProduct(U,Spring);
        }
    }
}   

/*----------------------------------------------------------------------------

ComputeJointForces - compute forces for the joints

*/

void SetSubMatrix( matrix &M, int unsigned Row, int unsigned Column,
                   matrix_3x3 const &Sub )
{
    for(int unsigned i = 0;i < 3;i++)
    {
        for(int unsigned j = 0;j < 3;j++)
        {
            M(Row+i,Column+j) = Sub.GetElement(i,j);
        }
    }
}
// @todo wrong hack!!!
void SetSubVector( matrix &M, int unsigned Row, 
                   vector_3 const &Sub )
{
    for(int unsigned i = 0;i < 3;i++)
    {
        M(Row+i,0) = Sub.GetElement(i);
    }
}
void AddSubMatrix( matrix &M, int unsigned Row, int unsigned Column,
                   matrix_3x3 const &Sub )
{
    for(int unsigned i = 0;i < 3;i++)
    {
        for(int unsigned j = 0;j < 3;j++)
        {
            M(Row+i,Column+j) += Sub.GetElement(i,j);
        }
    }
}
// @todo wrong hack!!!
void AddSubVector( matrix &M, int unsigned Row, 
                   vector_3 const &Sub )
{
    for(int unsigned i = 0;i < 3;i++)
    {
        M(Row+i,0) += Sub.GetElement(i);
    }
}
// @todo wrong hack!!!
vector_3 GetSubVector( matrix const &M, int unsigned Row )
{
    vector_3 v;
    for(int unsigned i = 0;i < 3;i++)
    {
        v.SetElement(i,M(Row+i,0));
    }
    return v;
}

void simulation_world::ComputeJointForces( int ConfigurationIndex )
{
    matrix C(NumberOfJoints*3), C_dot(NumberOfJoints*3);
    matrix Cq(NumberOfJoints*3,NumberOfBodies*6);
    matrix Cq_dotq_dot(NumberOfJoints*3);
    matrix MInverse(NumberOfBodies*6,NumberOfBodies*6);
    matrix MInverseFe_or_q_ddot(NumberOfBodies*6);
    int i, j;

    for(i = 0;i < NumberOfBodies;i++)
    {
        rigid_body &Body = aBodies[i];
        rigid_body::configuration &Configuration =
            Body.aConfigurations[ConfigurationIndex]; 

        matrix_3x3 M(Body.OneOverMass,matrix_3x3::Diagonal);
        matrix_3x3 const &I = Configuration.InverseWorldInertiaTensor;
        
        SetSubMatrix(MInverse,i*6  ,i*6  ,M);
        SetSubMatrix(MInverse,i*6+3,i*6+3,I);

        if(Body.IsKinematic)
        {
            SetSubVector(MInverseFe_or_q_ddot,i*6,
                         Configuration.CMAcceleration);
            SetSubVector(MInverseFe_or_q_ddot,i*6+3,
                         Configuration.AngularAcceleration);
        }
        else
        {
            // this body is dynamically simulated, the normal case
            SetSubVector(MInverseFe_or_q_ddot,i*6  ,M*Configuration.CMForce);
            SetSubVector(MInverseFe_or_q_ddot,i*6+3,
                         I*(Configuration.Torque -
                            CrossProduct(Configuration.AngularVelocity,
                                         Configuration.AngularMomentum)));
        }
    }

    for(i = 0;i < NumberOfJoints;i++)
    {
        for(j = 0;j < 2;j++)
        {
            spherical_joint::joint_definition &Joint = aJoints[i].aJointDefinitions[j];
            
            rigid_body &Body = aBodies[Joint.BodyIndex];
            rigid_body::configuration &Configuration =
                Body.aConfigurations[ConfigurationIndex];
            vector_3 const &omega = Configuration.AngularVelocity;
            vector_3 r = Configuration.Orientation * Joint.ToLocalJoint;
            vector_3 r_dot = CrossProduct(omega,r);
            vector_3 p = Configuration.CMPosition + r;
            vector_3 p_dot = Configuration.CMVelocity + r_dot;
                

            AddSubMatrix(Cq,i*3,Joint.BodyIndex*6,
                         j ?  matrix_3x3() :
                         -matrix_3x3());
            AddSubMatrix(Cq,i*3,Joint.BodyIndex*6+3,
                         j ? -matrix_3x3(r,matrix_3x3::SkewSymmetric) :
                         matrix_3x3(r,matrix_3x3::SkewSymmetric));

            AddSubVector(Cq_dotq_dot,i*3,
                         j ?  CrossProduct(omega,r_dot) :
                         -CrossProduct(omega,r_dot));

            AddSubVector(C,i*3,j ? p : -p);
            AddSubVector(C_dot,i*3,j ? p_dot : -p_dot);
        }
    }

    matrix CqT = Transpose(Cq);
    matrix A = Cq * MInverse * CqT;
    matrix x(NumberOfJoints*3);
    // @todo omigod
    real const OneOverTau = r(1) / r(0.125);
    real const Alpha = r(2) * OneOverTau, Beta = OneOverTau * OneOverTau;
    matrix Stabilizer(NumberOfJoints*3);
    if(StabilizersActive)
    {
        Stabilizer = Alpha * C_dot + Beta * C;
    }
    matrix b = (r(-1) * Cq_dotq_dot) - Cq * MInverseFe_or_q_ddot - Stabilizer;

    SolveLinearSystem(A,x,b);

    matrix Fc = CqT * x;

#if 0   
    PrintMatrix("Cq =",Cq);
    PrintMatrix("MInverse =",MInverse);
    PrintMatrix("A =",A);
    PrintMatrix("b =",b);
    PrintMatrix("MInverseFe_or_q_ddot =",MInverseFe_or_q_ddot);
    PrintMatrix("Cq_dotq_dot =",Cq_dotq_dot);
    PrintMatrix("x =",x);
    PrintMatrix("Fc =",Fc);
#endif  
    
    for(i = 0;i < NumberOfBodies;i++)
    {
        rigid_body &Body = aBodies[i];
        rigid_body::configuration &Configuration =
            Body.aConfigurations[ConfigurationIndex]; 

        Configuration.CMForce += GetSubVector(Fc,i*6);
        Configuration.Torque += GetSubVector(Fc,i*6+3);
    }

#if 0
    for(int i = 0;i < NumberOfJoints;i++)
    {
        spherical_joint::joint_definition &Inner = aJoints[i].aJointDefinitions[0];
        spherical_joint::joint_definition &Outer = aJoints[i].aJointDefinitions[1];

        rigid_body &BodyInner = aBodies[Inner.BodyIndex];
        rigid_body::configuration &ConfigurationInner =
            BodyInner.aConfigurations[ConfigurationIndex];
        vector_3 PositionInner =
            ConfigurationInner.CMPosition +
            ConfigurationInner.Orientation * Inner.ToLocalJoint;
        vector_3 UInner = PositionInner - ConfigurationInner.CMPosition;
        vector_3 VUInner = ConfigurationInner.CMVelocity +
            CrossProduct(ConfigurationInner.AngularVelocity,UInner);

        rigid_body &BodyOuter = aBodies[Outer.BodyIndex];
        rigid_body::configuration &ConfigurationOuter =
            BodyOuter.aConfigurations[ConfigurationIndex];
        vector_3 PositionOuter =
            ConfigurationOuter.CMPosition +
            ConfigurationOuter.Orientation * Outer.ToLocalJoint;
        vector_3 UOuter = PositionOuter - ConfigurationOuter.CMPosition;
        vector_3 VUOuter = ConfigurationOuter.CMVelocity +
            CrossProduct(ConfigurationOuter.AngularVelocity,UOuter);
    
        // spring goes from Inner to Outer
    
        vector_3 SpringVector = PositionOuter - PositionInner;
        vector_3 Spring = -Kbs * SpringVector * r(2);
    
        vector_3 RelativeVelocity = VUOuter - VUInner;
        // project velocity onto spring to get damping vector
        // this is basically a Gram-Schmidt projection
        vector_3 DampingForce;
        if(SpringVector.Length() > Epsilon)
        {
            DampingForce =
                -Kbd * (DotProduct(RelativeVelocity,SpringVector)/
                        DotProduct(SpringVector,SpringVector)) * SpringVector;
        }
    
        Spring += DampingForce;
    
        ConfigurationInner.CMForce -= Spring;
        ConfigurationInner.Torque -= CrossProduct(UInner,Spring);
    
        ConfigurationOuter.CMForce += Spring;
        ConfigurationOuter.Torque += CrossProduct(UOuter,Spring);
    }
#endif  
}   

/*----------------------------------------------------------------------------

CalculateVertices - figure out the body vertices from the configuration

@todo should't this be in the body?

*/

void simulation_world::CalculateVertices( int ConfigurationIndex )
{
    for(int Counter = 0;Counter < NumberOfBodies;Counter++)
    {
        rigid_body &Body = aBodies[Counter];
        rigid_body::configuration &Configuration =
            Body.aConfigurations[ConfigurationIndex];
            
        matrix_3x3 const &A = Configuration.Orientation;
        vector_3 const &R = Configuration.CMPosition;

        assert(Body.NumberOfBoundingVertices <
               Body.MaxNumberOfBoundingVertices);
        for(int unsigned i = 0;i < Body.NumberOfBoundingVertices;i++)
        {
            Configuration.aBoundingVertices[i] =
                R + A * Body.aBodyBoundingVertices[i];
        }
    }
}

/*----------------------------------------------------------------------------

Integrate - integrate the rigid body configurations forward in time from
                source config to target config

@todo calculate initial derivatives _once_
@todo use something better than Euler's method

*/

void simulation_world::Integrate( real DeltaTime )
{
    int Counter;

    for(Counter = 0;Counter < NumberOfBodies;Counter++)
    {
        rigid_body::configuration &Source =
            aBodies[Counter].aConfigurations[SourceConfigurationIndex];
        rigid_body::configuration &Target =
            aBodies[Counter].aConfigurations[TargetConfigurationIndex];

        // integrate primary quantities

        Target.CMPosition = Source.CMPosition +
            DeltaTime * Source.CMVelocity;

        Target.Orientation = Source.Orientation +
            DeltaTime *
            matrix_3x3(Source.AngularVelocity,matrix_3x3::SkewSymmetric) *
            Source.Orientation;

        if(aBodies[Counter].IsKinematic)
        {
            // @todo hack!
            Target.CMVelocity = Source.CMVelocity +
                DeltaTime * Source.CMAcceleration;
            Target.AngularVelocity = Source.AngularVelocity +
                DeltaTime * Source.AngularAcceleration;
            // @todo Target.AngularMomentum == INF?
        }
        else
        {
            Target.CMVelocity = Source.CMVelocity +
                (DeltaTime * aBodies[Counter].OneOverMass) * Source.CMForce;
            
            Target.AngularMomentum = Source.AngularMomentum +
                DeltaTime * Source.Torque;
        }

        OrthonormalizeOrientation(Target.Orientation);

        // compute auxiliary quantities

        Target.InverseWorldInertiaTensor = Target.Orientation *
            aBodies[Counter].InverseBodyInertiaTensor *
            Transpose(Target.Orientation);
        
        if(!aBodies[Counter].IsKinematic)
        {
            Target.AngularVelocity = Target.InverseWorldInertiaTensor *
                Target.AngularMomentum;
        }
    }
}


/*----------------------------------------------------------------------------

CheckForCollisions - test the current configuration for collisions

@todo handle multiple simultaneous collisions

*/

simulation_world::collision_state
simulation_world::CheckForCollisions( int ConfigurationIndex )
{
    // be optimistic!
    CollisionState = Clear;
    float const DepthEpsilon = 0.001f;

    for(int BodyIndex = 0;(BodyIndex < NumberOfBodies) &&
            (CollisionState != Penetrating);BodyIndex++)
    {
        rigid_body &Body = aBodies[BodyIndex];
        rigid_body::configuration &Configuration =
                Body.aConfigurations[ConfigurationIndex];

        for(int unsigned Counter = 0;
            (Counter < Body.NumberOfBoundingVertices) &&
                (CollisionState != Penetrating);Counter++)
        {
            vector_3 Position = Configuration.aBoundingVertices[Counter];
            vector_3 U = Position - Configuration.CMPosition;

            vector_3 Velocity = Configuration.CMVelocity +
                        CrossProduct(Configuration.AngularVelocity,U);

            for(int WallIndex = 0;(WallIndex < NumberOfWalls) &&
                (CollisionState != Penetrating);WallIndex++)
            {
                wall &Wall = aWalls[WallIndex];

                real axbyczd = DotProduct(Position,Wall.Normal) + Wall.d;

                if(axbyczd < -DepthEpsilon)
                {
                    CollisionState = Penetrating;
                }
                else
                if(axbyczd < DepthEpsilon)
                {
                    real RelativeVelocity = DotProduct(Wall.Normal,Velocity);

                    if(RelativeVelocity < r(0))
                    {
                        CollisionState = Colliding;
                        CollisionNormal = Wall.Normal;
                        CollidingCornerIndex = Counter;
                        CollidingBodyIndex = BodyIndex;
                    }
                }
            }
        }
    }

    return CollisionState;
}


/*----------------------------------------------------------------------------

ResolveCollisions - calculate collision impulses

@todo handle multiple simultaneous collisions

*/

void simulation_world::ResolveCollisions( int ConfigurationIndex )
{
    rigid_body &Body = aBodies[CollidingBodyIndex];
    rigid_body::configuration &Configuration =
        Body.aConfigurations[ConfigurationIndex];
    
    vector_3 Position =
        Configuration.aBoundingVertices[CollidingCornerIndex];
    
    vector_3 R = Position - Configuration.CMPosition;

    vector_3 Velocity = Configuration.CMVelocity +
        CrossProduct(Configuration.AngularVelocity,R);
    
    real ImpulseNumerator = -(r(1) + Body.CoefficientOfRestitution) *
        DotProduct(Velocity,CollisionNormal);

    real ImpulseDenominator = Body.OneOverMass +
        DotProduct(CrossProduct(Configuration.InverseWorldInertiaTensor *
                                CrossProduct(R,CollisionNormal),R),
                   CollisionNormal);
    
    vector_3 Impulse = (ImpulseNumerator/ImpulseDenominator) * CollisionNormal;
    
    // apply impulse to primary quantities
    Configuration.CMVelocity += Body.OneOverMass * Impulse;
    Configuration.AngularMomentum += CrossProduct(R,Impulse);
    
    // compute affected auxiliary quantities
    Configuration.AngularVelocity = Configuration.InverseWorldInertiaTensor *
        Configuration.AngularMomentum;
}

